#include "macros.h" // needs to be declared as first entry in .cpp or .h file!
#include <string>
#include <iostream>
using namespace std;
#include "AbstractExercise.h"
#include <ctime>
#include <cstdlib>
// ...

namespace {
    const int SIZE_X = 30;
    const int SIZE_Y = 20;
    const char ALIVE = '*';
    const char DEAD  = ' ';
    typedef char lifeMatrix_t[SIZE_X][SIZE_Y];
}

/**
 * &copy; Bernhard Wagner, bw@xmlizer.biz, 2003. 
 *
 * Models a "mod n set" (see e.g., John K. Truss "Discrete Mathematics for Computer Scientists", p.7)
 */
class ModN {
public:

	/**
	 * @param mod  The mod of the set  
	 * @param value  The value of the variable  
     */
    ModN(int mod, int value = 0) : fMod(mod), fValue(value), fCompleted(false) {}

	/**
	 * Asserts maintenance of affiliation to the set, i.e. "wraps around" 
	 * Side effect: Updates inner state of completed which is used for iteration.
	 */
    inline ModN& operator++(int) {
        fValue = (fValue + 1) % fMod;
        fCompleted = (fValue == 0);
        return *this;
    }

	/**
	 * Asserts maintenance of affiliation to the set, i.e. "wraps around" 
	 * Side effect: Updates inner state of completed which is used for iteration.
	 */
    inline ModN& operator--(int) {
        fValue = (fValue + fMod - 1) % fMod;
        fCompleted = (fValue == fMod - 1);
        return *this;
    }

	/**
	 * We compare mod and the value.
	 * FIXME: Probably should throw exception if mods are not the same.
	 */
    inline bool operator==(const ModN& operand) const { return (fMod == operand.fMod) && (fValue == operand.fValue); }

	/**
	 * As a convenience so ModN can be used directly within integer algebraic expressions. 
	 * There is a danger though, since comparisons might lead to confusing results e.g.
	 * <CODE><PRE>
	 * ModN mod34(4,3) int i = 4; mod43 == i; // ok, false, mod43 is 3, let's increment...
	 * mod43++; mod43 == i; // oops, false too, because of wrap around, i.e. mod43 now is 0!
	 * </PRE></CODE>
	 */
	inline operator int() const { return fValue; }

	/**
	 * Simulates iterator behaviour. Is set to true if the last increment or decrement 
	 * lead to wraparound.
	 *
	 * @return True if the last increment or decrement lead to wraparound.
	 */
    inline bool completed() { return fCompleted; }
private:
    int fMod;
    int fValue;
    bool fCompleted;
};

/**
 * LifeMatrix. 
 * &copy; Bernhard Wagner, bw@xmlizer.biz, 2003.
 */
class LifeMatrix {
public:
	/**
	 * Displays the LifeMatrix on stdout.
	 */
    inline void display() const {
        for(int i = 0; i<SIZE_Y; i++) {
            cout << "|";
            for(int j = 0; j<SIZE_X; j++){
                cout << fMatrix[j][i];
            }
            cout << "|";
            cout << endl;
        }   
        cout << endl;
    }

	/**
	 * Counts the neighbours of the cell given by x and y. 
	 * @param x xcoordinate of the cell of which to count the neighbours.
	 * @param y xcoordinate of the cell of which to count the neighbours.
	 * @return The number of neighbours of the cell given by x and y.
	 */
    inline int countNeighbours (const ModN& x, const ModN& y) const {
        int neighbours = 0;

        ModN mX = x;
        mX--;
        for(int i = 0; i<3; i++){ // can't use mX as counter because of mod n !
            ModN mY = y;
            mY--;
            for(int j = 0; j<3; j++){
                neighbours += !((mX == x)&&(mY == y)) && isAlive(mX, mY) ? 1 : 0;
                mY++;
            }
            mX++;
        }
        return neighbours;
    }

	/**
	 * Fills the LifeMatrix with a certain pattern. 
	 * You can choose between fillRand, fillSquare58, fillLine, fillGlider
	 */
    inline void fill(){
        for(int i = 0; i<SIZE_X; i++){
            for(int j = 0; j<SIZE_Y; j++){
                fMatrix[i][j] = fillGlider(i, j);
            }
        }   
    }

	/** 
	 * Kills the element at the given corodinates.
	 * @param x xcoordinate of the cell to kill.
	 * @param y ycoordinate of the cell to kill.
	 */
    inline void kill(const ModN& i, const ModN& j){
        fMatrix[i][j] = DEAD;
    }

	/** 
	 * Enlivens the element at the given corodinates.
	 * @param x xcoordinate of the cell to enliven.
	 * @param y ycoordinate of the cell to enliven.
	 */
    inline void enliven(const ModN& x, const ModN& y){
        fMatrix[x][y] = ALIVE;
    }

	/** 
	 * Indicates whether the element at the given corodinates is alive.
	 * @param x xcoordinate of the cell to check.
	 * @param y ycoordinate of the cell to check.
	 */
    inline bool isAlive(const ModN& x, const ModN& y) const {
        return fMatrix[x][y] == ALIVE;
    }

	/** 
	 * Seeds the random number generator.
	 */
    static inline void initRand() {
        long sek;
        time(&sek);
        srand((unsigned)sek);
    }

private:

	/**
	 * Fills the LifeMatrix with a random pattern.
	 */
    inline char fillRand(int x, int y) {
        return rand()%2 ? ALIVE : DEAD;
    }

	/**
	 * Fills the LifeMatrix with a square pattern.
	 */
    inline char fillSquare58(int x, int y){
        return (x>5)&&(x<8)&&(y>5)&&(y<8) ? ALIVE : DEAD;
    }

	/**
	 * Fills the LifeMatrix with a line pattern.
	 */
    inline char fillLine(int x, int y){
        return (x==8 || x== 12)&&(y>5)&&(y<9)  ? ALIVE : DEAD;
    }

	/**
	 * Fills the LifeMatrix with the Glider pattern.
	 */
    inline char fillGlider(int x, int y){
        return 
            (x==0 && y== 0)
            ||(x==0 && y== 1)
            ||(x==0 && y== 2)
            ||(x==1 && y== 0)
            ||(x==2 && y== 1)
            ? ALIVE : DEAD;
    }

    lifeMatrix_t fMatrix;
};

/**
 * Game of Life invented by John Horton Conway 
 * &copy; Bernhard Wagner, bw@xmlizer.biz, 2003.
 */
class Life {
public:
    Life() : fNow(new LifeMatrix()), fFuture(new LifeMatrix()) {}

	/**
	 * Runs the game.
	 */
    inline void run(){
        LifeMatrix::initRand();
        fill();
        while(true){
            display();
            calculate();
            wait();
        }
    }

	~Life(){
		delete fNow;
		delete fFuture;
	}
private:

	/**
	 * Fills the LifeMatrix with initial setting. 
	 */
    inline void Life::fill(){
        fNow->fill();
    }

	/**
	 * Clears screen and displays the current LifeMatrix on stdout.
	 */
    inline void Life::display(){
        cls();
        fNow->display();
    }

	/**
	 * Death if an occupied cell has 0, 1, 4, 5, 6, 7, or 8 occupied neighbours,
	 * the organism dies (0, 1 neighbors: of loneliness; 4 thru 8: of overcrowding).
	 * Survival if an occupied cell has two or three neighbours, the organism survives 
	 * to the next generation. Birth If an unoccupied cell has three occupied neighbors, 
	 * it becomes occupied.
	 */
    inline void calculate(){
        for(ModN i(SIZE_X, 0); !i.completed(); i++){
            for(ModN j(SIZE_Y, 0); !j.completed(); j++){
                int n = fNow->countNeighbours(i, j);
                if((n < 2)||(n > 3)){
                    fFuture->kill(i, j);
                } else if ((n == 2) && (fNow->isAlive(i, j))) {
                    fFuture->enliven(i, j);
                } else if (n == 3){
                    fFuture->enliven(i, j);
                } else {
                    fFuture->kill(i, j);
                }
            }
        }
        swap();
    }

	/**
	 * Swap future and now. Comparable to double buffering for raster graphics.
	 */
    inline void Life::swap(){
        LifeMatrix* tmp = fNow;
        fNow = fFuture;
        fFuture = tmp;
    }

	/**
	 * Busy wait.
	 */
    static inline void Life::wait() {
        for(long i= 0;i<19999999L;i++){}
    }

	/**
	 * Wrapper for clear screen functionality.
	 */
    static inline void Life::cls(){
        system("cls");
    }

    LifeMatrix* fNow;
    LifeMatrix* fFuture;
};

/**
 * The LifeExercise.
 * &copy; Bernhard Wagner, bw@xmlizer.biz, 2003.
 */
class LifeExercise : public AbstractExercise {
public:

    LifeExercise(string name) : AbstractExercise(name){}

protected:

    void execute();

private:
    Life fLife;
};
    
namespace {
    LifeExercise proto("LifeExercise");
}

void LifeExercise::execute(){
    fLife.run();
}


